"""
credit : https://github.com/ludvb/batchrenorm/blob/master/batchrenorm/batchrenorm.py
"""

import torch


# __all__ = ["BatchRenorm1d", "BatchRenorm2d", "BatchRenorm3d"]


class BatchRenorm(torch.nn.Module):
    def __init__(
        self,
        num_features: int,
        eps: float = 1e-3,
        momentum: float = 0.01,
        affine: bool = True,
    ):
        super(BatchRenorm,self).__init__()
        self.register_buffer(
            "running_mean", torch.zeros(num_features, dtype=torch.float)
        )
        self.register_buffer(
            "running_std", torch.ones(num_features, dtype=torch.float)
        )
        self.register_buffer(
            "num_batches_tracked", torch.tensor(0, dtype=torch.long)
        )
        self.weight = torch.nn.Parameter(
            torch.ones(num_features, dtype=torch.float)
        )
        self.bias = torch.nn.Parameter(
            torch.zeros(num_features, dtype=torch.float)
        )
        self.affine = affine
        self.eps = eps
        self.step = 0
        self.momentum = momentum

    def _check_input_dim(self, x: torch.Tensor) -> None:
        raise NotImplementedError()  # pragma: no cover

    @property
    def rmax(self) -> torch.Tensor:
        return (2 / 35000 * self.num_batches_tracked + 25 / 35).clamp_(
            1.0, 3.0
        )

    @property
    def dmax(self) -> torch.Tensor:
        return (5 / 20000 * self.num_batches_tracked - 25 / 20).clamp_(
            0.0, 5.0
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        self._check_input_dim(x)
        if x.dim() > 2:
            x = x.transpose(1, -1)
        if self.training:
            dims = [i for i in range(x.dim() - 1)]
            batch_mean = x.mean(dims)
            batch_std = x.std(dims, unbiased=False) + self.eps
            r = (
                batch_std.detach() / self.running_std.view_as(batch_std)
            ).clamp_(1 / self.rmax, self.rmax)
            d = (
                (batch_mean.detach() - self.running_mean.view_as(batch_mean))
                / self.running_std.view_as(batch_std)
            ).clamp_(-self.dmax, self.dmax)
            x = (x - batch_mean) / batch_std * r + d
            self.running_mean += self.momentum * (
                batch_mean.detach() - self.running_mean
            )
            self.running_std += self.momentum * (
                batch_std.detach() - self.running_std
            )
            self.num_batches_tracked += 1
        else:
            x = (x - self.running_mean) / self.running_std
        if self.affine:
            x = self.weight * x + self.bias
        if x.dim() > 2:
            x = x.transpose(1, -1)
        return x


class BatchRenorm1d(BatchRenorm):
    def _check_input_dim(self, x: torch.Tensor) -> None:
        if x.dim() not in [2, 3]:
            raise ValueError("expected 2D or 3D input (got {x.dim()}D input)")


class BatchRenorm2d(BatchRenorm):
    def _check_input_dim(self, x: torch.Tensor) -> None:
        if x.dim() != 4:
            raise ValueError("expected 4D input (got {x.dim()}D input)")


class BatchRenorm3d(BatchRenorm):
    def _check_input_dim(self, x: torch.Tensor) -> None:
        if x.dim() != 5:
            raise ValueError("expected 5D input (got {x.dim()}D input)")


import torch
from torch import nn


class ShellProvider(nn.Module):
    """
    This layer calculates distance of each atom in a molecule to its
    closest neighbouring atoms.

    Based on the SchnetPack AtomDistances: https://github.com/atomistic-machine-learning/schnetpack under the MIT License.


    Parameters
    ----------

    """
    def __init__(self,
                 return_vecs=False,
                 normalize_vecs=False,
                 pbc=False,
                 cutoff=None):
        super(ShellProvider, self).__init__()
        self.return_vecs = return_vecs
        self.normalize_vecs = normalize_vecs
        self.epsilon = 1e-8
        self.pbc = pbc
        self.cutoff = cutoff

    def forward(self,
                atoms,
                neighbors,
                neighbor_mask=None,
                lattice=None):
        """
        The main driver to calculate distances of atoms in a shell from center atom.

        Parameters
        ----------
        atoms: torch.Tensor
            XYZ coordinates of atoms in molecules.
            shape: (B, A, 3)

        neighbors: torch.Tensor
            indices of adjacent atoms.
            shape: (B, A, N), for small systems N=A-1

        neighbor_mask: torch.tensor
            boolean mask for neighbor positions.

        lattice: torch.Tensor or None
            the pbc lattice array for current batch of data
            shape: (9,)


        Returns
        -------
        torch.Tensor: distances with shape: (B, A, N)
        torch.Tensor: distance vector with shape: (B, A, N, 3)

        Notes
        -----
        shape of tensors are specified with following symbols throughout the documentation:
            - B: batch size
            - A: max number of atoms
            - N: max number of neighbors (upper limit is A-1)

        """
        # Construct auxiliary index vector
        B, A, _ = atoms.size()
        idx_m = torch.arange(B, device=atoms.device, dtype=torch.long)[
                :, None, None
                ]

        # Get atomic positions of all neighboring indices
        ngh_atoms_xyz = atoms[idx_m, neighbors[:, :, :], :]

        # Subtract positions of central atoms to get distance vectors
        distance_vector = ngh_atoms_xyz - atoms[:, :, None, :]


        # pbc: for distance in a direction (d) and boxlength (L), d = (d + L/2) % L - L/2
        if self.pbc:
            lattice_shift_arr = torch.tensor([[i, j, k] for i in [-1, 0, 1] for j in [-1, 0, 1] for k in [-1, 0, 1]],
                                             dtype=lattice.dtype,
                                             device=lattice.device)  # 27 x 3
            lattice_batchlast = lattice.view((-1, 3, 3)).moveaxis(0, 2) # 3 x 3 x B
            distance_shift_arr = torch.tensordot(lattice_shift_arr, lattice_batchlast, 1).moveaxis(2, 1) # 27 x B x 3
            distance_vector_pbc = distance_vector[None] + distance_shift_arr[:, :, None, None] # 27 x B x A x N x 3
            distance_vector = distance_vector_pbc.moveaxis(0, -2).flatten(start_dim=2, end_dim=3) # B x A x N*27 x 3
            distances = torch.linalg.norm(distance_vector, dim=-1)

            # expand neighbor (and neighbor mask)
            neighbors = neighbors[:, :, :, None].tile((1, 1, 1, 27)).flatten(start_dim=2)  # B x A x Nx27
            if neighbor_mask is not None:
                neighbor_mask = neighbor_mask[:, :, :, None].tile((1, 1, 1, 27)).flatten(start_dim=2)
            # distance_min_idx = torch.argmin(distances_pbc, dim=0)
            # distance_min_idx_tiled = distance_min_idx[None, ..., None].tile((1, 1, 1, 1, 3))
            # distance_vector = torch.gather(distance_vector_pbc, 0, distance_min_idx_tiled).squeeze(0)
            # distances = torch.gather(distances_pbc, 0, distance_min_idx[None]).squeeze(0)


        #     x_vec = distance_vector[:, :, :, 0]
        #     x_vec = (x_vec + 0.5 * self.box[0]) % self.box[0] - 0.5 * self.box[0]
        #     y_vec = distance_vector[:, :, :, 1]
        #     y_vec = (y_vec + 0.5 * self.box[1]) % self.box[1] - 0.5 * self.box[1]
        #     z_vec = distance_vector[:, :, :, 2]
        #     z_vec = (z_vec + 0.5 * self.box[2]) % self.box[2] - 0.5 * self.box[2]
        #     distance_vector[:, :, :, 0] = x_vec
        #     distance_vector[:, :, :, 1] = y_vec
        #     distance_vector[:, :, :, 2] = z_vec
        # # print(distance_vector)
        else:
            distances = torch.norm(distance_vector, 2, 3)   # B, A, N

        if neighbor_mask is not None:
            # Avoid problems with zero distances in forces (instability of square
            # root derivative at 0) This way is neccessary, as gradients do not
            # work with inplace operations, such as e.g.
            # -> distances[mask==0] = 0.0
            tmp_dist = torch.zeros_like(distances)
            tmp_dist[neighbor_mask != 0] = distances[neighbor_mask != 0]
            distances = tmp_dist
        # print(distances)

        if self.cutoff is not None:
            # remove all neighbors beyond cutoff to save computation
            within_cutoff = distances < self.cutoff
            if neighbor_mask is not None:
                within_cutoff[neighbor_mask == 0] = False
            neighbor_counts = torch.zeros((B, A), dtype=int)
            temporal_distances = [[[] for _ in range(A)] for _ in range(B)]
            temporal_distance_vec = [[[] for _ in range(A)] for _ in range(B)]
            temporal_neighbor = [[[] for _ in range(A)] for _ in range(B)]
            temporal_neighbor_mask = [[[] for _ in range(A)] for _ in range(B)]
            for i in range(B):
                for j in range(A):
                    neighbor_count = within_cutoff[i, j].sum()
                    neighbor_counts[i, j] = neighbor_count
                    temporal_distances[i][j] = distances[i, j, within_cutoff[i, j]]
                    temporal_distance_vec[i][j] = distance_vector[i, j, within_cutoff[i, j]]
                    temporal_neighbor[i][j] = neighbors[i, j, within_cutoff[i, j]]
                    temporal_neighbor_mask[i][j] = torch.tensor([1] * neighbor_count)
            N = neighbor_counts.max()
            distances = torch.zeros((B, A, N), device=atoms.device)
            distance_vector = torch.zeros((B, A, N, 3), device=atoms.device)
            neighbors = torch.zeros((B, A, N), device=atoms.device, dtype=torch.int64)
            neighbor_mask = torch.zeros((B, A, N), device=atoms.device)
            for i in range(B):
                for j in range(A):
                    distances[i, j, :neighbor_counts[i, j]] = temporal_distances[i][j]
                    distance_vector[i, j, :neighbor_counts[i, j]] = temporal_distance_vec[i][j]
                    neighbors[i, j, :neighbor_counts[i, j]] = temporal_neighbor[i][j]
                    neighbor_mask[i, j, :neighbor_counts[i, j]] = temporal_neighbor_mask[i][j]

        if self.return_vecs:
            tmp_distances = torch.ones_like(distances)
            tmp_distances[neighbor_mask != 0] = distances[neighbor_mask != 0] + self.epsilon

            if self.normalize_vecs:
                distance_vector = distance_vector / tmp_distances[:, :, :, None]
            return distances, distance_vector, neighbors, neighbor_mask

        return distances, neighbors, neighbor_mask


import numpy as np
from ase.units import *
from ase.calculators.calculator import Calculator
import torch
import torch.autograd.functional as F
import yaml

from newtonnet.layers.activations import get_activation_by_string
from newtonnet.models import NewtonNet
from newtonnet.data import ExtensiveEnvironment
from newtonnet.data import batch_dataset_converter


##-------------------------------------
##     ML model ASE interface
##--------------------------------------
class MLAseCalculator(Calculator):
    implemented_properties = ['energy', 'forces', 'hessian']

    ### Constructor ###
    def __init__(self, model_path, settings_path, hess_method=None, hess_precision=None, disagreement='std', device='cpu', **kwargs):
        """
        Constructor for MLAseCalculator

        Parameters
        ----------
        model_path: str or list of str
            path to the model. eg. '5k/models/best_model_state.tar'
        settings_path: str or list of str
            path to the .yml setting path. eg. '5k/run_scripts/config_h2.yml'
        hess_method: str
            method to calculate hessians. 
            None: do not calculate hessian (default)
            'autograd': automatic differentiation
            'fwd_diff': forward difference
            'cnt_diff': central difference
        hess_precision: float
            hessian gradient calculation precision for 'fwd_diff' and 'cnt_diff', ignored otherwise (default: None)
        disagreement: str
            method to calculate disagreement between models.
            'std': standard deviation of all votes (default)
            'std_outlierremoval': standard deviation with outlier removal
            'range': range of all votes
            'values': values of each vote
            None: do not calculate disagreement
        device: 
            device to run model. eg. 'cpu', ['cuda:0', 'cuda:1']
        kwargs
        """
        Calculator.__init__(self, **kwargs)

        if type(device) is list:
            self.device = [torch.device(item) for item in device]
        else:
            self.device = [torch.device(device)]

        self.hess_method = hess_method
        if self.hess_method == 'autograd':
            self.return_hessian = True
        elif self.hess_method == 'fwd_diff' or self.hess_method == 'cnt_diff':
            self.return_hessian = False
            self.hess_precision = hess_precision
        else:
            self.return_hessian = False

        self.disagreement = disagreement

        torch.set_default_tensor_type(torch.DoubleTensor)
        if type(model_path) is list:
            self.models = [self.load_model(model_path_, settings_path_) for model_path_, settings_path_ in zip(model_path, settings_path)]
        else:
            self.models = [self.load_model(model_path, settings_path)]
        

    def calculate(self, atoms=None, properties=['energy','forces','hessian'], system_changes=None):
        super().calculate(atoms,properties,system_changes)
        data = self.extensive_data_loader(data=self.data_formatter(atoms), device=self.device[0])
        energy = np.zeros((len(self.models), 1))
        forces = np.zeros((len(self.models), data['R'].shape[1], 3))
        hessian = np.zeros((len(self.models), data['R'].shape[1], 3, data['R'].shape[1], 3))
        if self.hess_method=='autograd':
            for model_, model in enumerate(self.models):
                #pred_E = lambda R: self.model(dict(data, R=R))
                #pred_F = torch.func.jacrev(pred_E)
                #pred_H = torch.func.hessian(pred_E)
                #energy = pred_E(data['R'])
                #forces = -pred_F(data['R'])
                #hessian = pred_H(data['R'])
                #energy = self.model(data).detach().cpu().numpy()
                #forces = -F.jacobian(lambda R: self.model(dict(data, R=R)), data['R']).detach().cpu().numpy()
                #hessian = F.hessian(lambda R: self.model(dict(data, R=R), vectorize=True), data['R']).detach().cpu().numpy()
                pred = model(data)
                energy[model_] = pred['E'].detach().cpu().numpy() * (kcal/mol)
                forces[model_] = pred['F'].detach().cpu().numpy() * (kcal/mol/Ang)
                hessian[model_] = pred['H'].detach().cpu().numpy() * (kcal/mol/Ang/Ang)
                del pred
        elif self.hess_method=='fwd_diff':
            for model_, model in enumerate(self.models):
                pred = model(data)
                energy[model_] = pred['E'].detach().cpu().numpy()[0] * (kcal/mol)
                forces_temp = pred['F'].detach().cpu().numpy() * (kcal/mol/Ang)
                forces[model_] = forces_temp[0]
                n = 1
                for A_ in range(data['R'].shape[1]):
                    for X_ in range(3):
                        hessian[model_, A_, X_, :, :] = -(forces_temp[n] - forces_temp[0]) / self.hess_precision
                        n += 1
                del pred
        elif self.hess_method=='cnt_diff':
            for model_, model in enumerate(self.models):
                pred = model(data)
                energy[model_] = pred['E'].detach().cpu().numpy()[0] * (kcal/mol)
                forces_temp = pred['F'].detach().cpu().numpy() * (kcal/mol/Ang)
                forces[model_] = forces_temp[0]
                n = 1
                for A_ in range(data['R'].shape[1]):
                    for X_ in range(3):
                        hessian[model_, A_, X_, :, :] = -(forces_temp[n] - forces_temp[n+1]) / 2 / self.hess_precision
                        n += 2
                del pred
        elif self.hess_method is None:
            for model_, model in enumerate(self.models):
                pred = model(data)
                energy[model_] = pred['E'].detach().cpu().numpy()[0] * (kcal/mol)
                forces[model_] = pred['F'].detach().cpu().numpy()[0] * (kcal/mol/Ang)
                del pred
        self.results['outlier'] = self.q_test(energy)
        self.results['energy'] = self.remove_outlier(energy, self.results['outlier']).mean()
        self.results['forces'] = self.remove_outlier(forces, self.results['outlier']).mean(axis=0)
        self.results['hessian'] = self.remove_outlier(hessian, self.results['outlier']).mean(axis=0)
        if self.disagreement=='std':
            self.results['energy_disagreement'] = energy.std()
            self.results['forces_disagreement'] = forces.std(axis=0).max()
            self.results['hessian_disagreement'] = hessian.std(axis=0).max()
        elif self.disagreement=='std_outlierremoval':
            self.results['energy_disagreement'] = self.remove_outlier(energy, self.results['outlier']).std()
            self.results['forces_disagreement'] = self.remove_outlier(forces, self.results['outlier']).std(axis=0).max()
            self.results['hessian_disagreement'] = self.remove_outlier(hessian, self.results['outlier']).std(axis=0).max()
        elif self.disagreement=='range':
            self.results['energy_disagreement'] = (energy.max() - energy.min())
            self.results['forces_disagreement'] = (forces.max(axis=0) - forces.min(axis=0)).max()
            self.results['hessian_disagreement'] = (hessian.max(axis=0) - hessian.min(axis=0)).max()
        elif self.disagreement=='values':
            self.results['energy_disagreement'] = energy
            self.results['forces_disagreement'] = forces
            self.results['hessian_disagreement'] = hessian
        del energy, forces, hessian


    def load_model(self, model_path, settings_path):
        settings = yaml.safe_load(open(settings_path, "r"))
        activation = get_activation_by_string(settings['model']['activation'])
        model = NewtonNet(resolution=settings['model']['resolution'],
                            n_features=settings['model']['n_features'],
                            activation=activation,
                            n_interactions=settings['model']['n_interactions'],
                            dropout=settings['training']['dropout'],
                            max_z=10,
                            cutoff=settings['data']['cutoff'],  ## data cutoff
                            cutoff_network=settings['model']['cutoff_network'],
                            normalize_atomic=settings['model']['normalize_atomic'],
                            requires_dr=settings['model']['requires_dr'],
                            device=self.device[0],
                            create_graph=False,
                            shared_interactions=settings['model']['shared_interactions'],
                            return_hessian=self.return_hessian,
                            double_update_latent=settings['model']['double_update_latent'],
                            layer_norm=settings['model']['layer_norm'],
                            )

        model.load_state_dict(torch.load(model_path, map_location=self.device[0])['model_state_dict'], )
        model = model
        model.to(self.device[0])
        model.eval()
        return model
    

    def data_formatter(self, atoms):
        """
        convert ase.Atoms to input format of the model

        Parameters
        ----------
        atoms: ase.Atoms

        Returns
        -------
        data: dict
            dictionary of arrays with following keys:
                - 'R':positions
                - 'Z':atomic_numbers
                - 'E':energy
                - 'F':forces
        """
        data  = {
            'R': np.array(atoms.get_positions())[np.newaxis, ...], #shape(ndata,natoms,3)
            'Z': np.array(atoms.get_atomic_numbers())[np.newaxis, ...], #shape(ndata,natoms)
            'E': np.zeros((1,1)), #shape(ndata,1)
            'F': np.zeros((1,len(atoms.get_atomic_numbers()), 3)),#shape(ndata,natoms,3)
        }
        if self.hess_method=='fwd_diff':
            n = data['R'].size
            data['R'] = np.tile(data['R'], (1 + n, 1, 1))
            data['Z'] = np.tile(data['Z'], (1 + n, 1))
            data['E'] = np.tile(data['E'], (1 + n, 1))
            data['F'] = np.tile(data['F'], (1 + n, 1, 1))
            n = 1
            for A_ in range(data['R'].shape[1]):
                for X_ in range(3):
                    data['R'][n, A_, X_] += self.hess_precision
                    n += 1
        if self.hess_method=='cnt_diff':
            n = data['R'].size
            data['R'] = np.tile(data['R'], (1 + 2*n, 1, 1))
            data['Z'] = np.tile(data['Z'], (1 + 2*n, 1))
            data['E'] = np.tile(data['E'], (1 + 2*n, 1))
            data['F'] = np.tile(data['F'], (1 + 2*n, 1, 1))
            n = 1
            for A_ in range(data['R'].shape[1]):
                for X_ in range(3):
                    data['R'][n, A_, X_] += self.hess_precision
                    data['R'][n+1, A_, X_] -= self.hess_precision
                    n += 2
        return data


    def extensive_data_loader(self, data, device=None):
        batch = {'R': data['R'], 'Z': data['Z']}
        N, NM, AM, _, _ = ExtensiveEnvironment().get_environment(data['R'], data['Z'])
        batch.update({'N': N, 'NM': NM, 'AM': AM})
        batch = batch_dataset_converter(batch, device=device)
        return batch
    

    def remove_outlier(self, data, idx):
        if idx is None:
            return data
        else:
            return np.delete(data, idx, axis=0)


    def q_test(self, data):
        """
        Dixon's Q test for outlier detection

        Parameters
        ----------
        data: 1d array with shape (nlearners,)

        Returns
        -------
        idx: int or None
            the index to be filtered out (return only one index for now as the default is only 4 learners)
        """
        if len(data) < 3:
            idx = None
        else:
            q_ref = { 3: 0.970,  4: 0.829,  5: 0.710, 
                      6: 0.625,  7: 0.568,  8: 0.526,  9: 0.493, 10: 0.466, 
                     11: 0.444, 12: 0.426, 13: 0.410, 14: 0.396, 15: 0.384, 
                     16: 0.374, 17: 0.365, 18: 0.356, 19: 0.349, 20: 0.342, 
                     21: 0.337, 22: 0.331, 23: 0.326, 24: 0.321, 25: 0.317, 
                     26: 0.312, 27: 0.308, 28: 0.305, 29: 0.301, 30: 0.290}.get(len(self.models))  # 95% confidence interval
            sorted_data = np.sort(data, axis=0)
            q_stat_min = (sorted_data[1] - sorted_data[0]) / (sorted_data[-1] - sorted_data[0])
            q_stat_max = (sorted_data[-1] - sorted_data[-2]) / (sorted_data[-1] - sorted_data[0])
            if q_stat_min > q_ref:
                idx = np.argmin(data)
            elif q_stat_max > q_ref:
                idx = np.argmax(data)
            else:
                idx = None
        return idx


# ##-------------------------------------
# ##     ASE interface for Plumed calculator
# ##--------------------------------------
# class PlumedCalculator(Calculator):
#     implemented_properties = ['energy', 'forces']  # , 'stress'
#     def __init__(self, ase_plumed, **kwargs):
#         Calculator.__init__(self, **kwargs)
#         self.ase_plumed = ase_plumed
#         self.counter = 0
#         self.prev_force =None
#         self.prev_energy = None

#     def calculate(self, atoms=None, properties=['forces'],system_changes=None):
#         super().calculate(atoms,properties,system_changes)
#         forces = np.zeros((atoms.get_positions()).shape)
#         energy = 0

#         model_force = np.copy(forces)
#         self.counter += 1
#         # every step() call will call get_forces 2 times, only do plumed once(2nd) to make metadynamics work correctly
#         # there is one call to get_forces when initialize
#         # print(self.counter)
#         # plumed_forces, plumed_energy = self.ase_plumed.external_forces(self.counter , new_forces=forces,
#         #
#         #                                                                delta_forces=True)
#         if self.counter % 2 == 1:
#             plumed_forces,plumed_energy = self.ase_plumed.external_forces((self.counter + 1) // 2 - 1, new_forces=forces,
#                                                             new_energy=energy,delta_forces=True)
#             self.prev_force = plumed_forces
#             self.prev_energy = plumed_energy
#             # print('force diff', np.sum(plumed_forces - model_force))
#         else:
#             plumed_forces = self.prev_force
#             plumed_energy = self.prev_energy
#             # print(self.counter)
#         # if self.counter % 500 == 0:
#         #     print('force diff', np.linalg.norm(plumed_forces - model_force))


#         # delta energy and forces
#         if 'energy' in properties:
#             self.results['energy'] = plumed_energy
#         if 'forces' in properties:
#             self.results['forces'] = plumed_forces

# if __name__ == '__main__':
#     pass



import torch
from torch import nn
import numpy as np


def get_activation_by_string(key):
    if key == "swish":
        activation = swish
    elif key == 'relu':
        activation = nn.ReLU()
    elif key == 'ssp':
        activation = shifted_softplus
    elif key == 'gelu':
        activation = gelu
    else:
        raise NotImplementedError("The activation function '%s' is unknown."%str(key))
    return activation


def swish(x):
    r"""Compute the self-gated Swish activation function.

    .. math::
       y = x * sigmoid(x)

    Args:
        x (torch.Tensor): input tensor.

    Returns:
        torch.Tensor: Swish activation of input.

    """
    return x * torch.sigmoid(x)


def shifted_softplus(x):
    r"""Compute shifted soft-plus activation function.
    Copied from: https://github.com/atomistic-machine-learning/schnetpack under the MIT License.

    .. math::
       y = \ln\left(1 + e^{-x}\right) - \ln(2)

    Args:
        x (torch.Tensor): input tensor.

    Returns:
        torch.Tensor: shifted soft-plus of input.

    """
    return nn.functional.softplus(x) - np.log(2.0)


def gelu(x):
    r"""Compute the gaussian error linear unit (GELU) activation function.

    .. math::
       y = x * sigmoid(1.702x)

    Args:
        x (torch.Tensor): input tensor.

    Returns:
        torch.Tensor: Swish activation of input.

    """
    return x * torch.sigmoid(1.702*x)


import setuptools
from os import path
import newtonnet

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md')) as f:
    long_description = f.read()

if __name__ == "__main__":
    setuptools.setup(
        name=newtonnet.__name__,
        version=newtonnet.__version__,
        author='Teresa Head-Gordon',
        author_email='thg@berkeley.edu',
        project_urls={
            'Source': 'https://github.com/THGLab/NewtonNet',
        },
        description=
        "A Newtonian message passing network for deep learning of interatomic potentials and forces",
        long_description=long_description,
        long_description_content_type="text/markdown",
        scripts=['cli/newtonnet_train'],
        keywords=[
            'Machine Learning', 'Data Mining', 'Quantum Chemistry',
            'Molecular Dynamics'
        ],
        license='MIT',
        packages=setuptools.find_packages(),
        install_requires=["numpy", "scipy", "scikit-learn", "pandas", "ase", "tqdm", "pyyaml", "torch"],
        include_package_data=True,
        classifiers=[
            'Development Status :: 4 - Beta',
            'Natural Language :: English',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3',
        ],
        zip_safe=False,
    )



import numpy as np
import torch
from torch import nn


class PolynomialCutoff(nn.Module):
    r"""Class of Polynomial cutoff.

    credit : Directional Message Passing for Molecular Graphs, Johannes Klicpera, Janek Grob, Stephan Gunnemann
             Published at ICLR 2020


    Parameters
    ----------
    cutoff (float, optional): cutoff radius.

    """
    def __init__(self, cutoff=5.0, p=9):
        super(PolynomialCutoff, self).__init__()
        self.register_buffer("cutoff", torch.FloatTensor([cutoff]))
        self.register_buffer("p", torch.FloatTensor([p]))

    def forward(self, distances):
        """Compute cutoff.

        Args:
            distances (torch.Tensor): values of interatomic distances.

        Returns:
            torch.Tensor: values of cutoff function.

        """
        d = distances / self.cutoff

        # Compute values of cutoff function
        cutoffs = 1 - 0.5*(self.p+1)*(self.p+2)*d.pow(self.p) + self.p*(self.p+2)*d.pow(self.p+1) - 0.5*self.p*(self.p+1)*d.pow(self.p+2)

        # Remove contributions beyond the cutoff radius
        cutoffs *= (distances < self.cutoff).float()

        return cutoffs


class CosineCutoff(nn.Module):
    r"""Class of Behler cosine cutoff.

    credit : https://github.com/atomistic-machine-learning/schnetpack under the MIT License

    .. math::
       f(r) = \begin{cases}
        0.5 \times \left[1 + \cos\left(\frac{\pi r}{r_\text{cutoff}}\right)\right]
          & r < r_\text{cutoff} \\
        0 & r \geqslant r_\text{cutoff} \\
        \end{cases}

    Args:
        cutoff (float, optional): cutoff radius.

    """
    def __init__(self, cutoff=5.0):
        super(CosineCutoff, self).__init__()
        self.register_buffer("cutoff", torch.FloatTensor([cutoff]))

    def forward(self, distances):
        """Compute cutoff.

        Args:
            distances (torch.Tensor): values of interatomic distances.

        Returns:
            torch.Tensor: values of cutoff function.

        """
        # Compute values of cutoff function
        cutoffs = 0.5 * (torch.cos(distances * np.pi / self.cutoff) + 1.0)
        # Remove contributions beyond the cutoff radius
        cutoffs *= (distances < self.cutoff).float()
        return cutoffs
